﻿//----- Refer To Steam Workshop Discussion Section For Variables Definition -----

//Configures the missile launch mode to use when script timer loop is activated:
//
//  0 = Launch missile in Active Radar Homing Mode using launching ship's R_REMOTE for initial lock-on. Switches over to on-board Radar when lock-on is stable.
//  1 = Launch missile in Active Radar Homing Mode using coordinates from launching ship's R_TARGET for initial lock-on. Switches over to on-board Radar when lock-on is stable.
//  2 = Launch missile in Semi-Active Radar Homing Mode using launching ship's R_REMOTE as the lock-on Radar permanently.
//  3 = Launch missile in Camera Guided Mode using launching ship's R_CAMERA for direction tracking.
//  4 = Launch missile in Cruise Mode using launching ship's R_TARGET as GPS coordinates.
//  5 = Launch missile in Active Radar Homing Mode using launching ship's R_CAMERA for initial lock-on. This initial lock-on takes longer than mode 0 and 1 as the missile's perspective is different.
//  6 = Launch missile in Active Radar Homing Mode using launching ship's R_REMOTE for initial lock-on with R_CAMERA for aiming.
//  80 = Launch missile in Evil NPC Mode (Only works for NPC ships). Homes in on the nearest player.
//  99 = Launch missile in Dummy Mode. This is used for testing if the missile detaches properly at the correct connection and display missile detected configurations.
//
int missileLaunchType = 0;

//Type of block to disconnect missile from launching ship: 0 = Merge Block, 1 = Rotor, 2 = Connector, 3 = Merge Block And Any Locked Connectors, 4 = Rotor And Any Locked Connectors, 99 = No detach required
int missileDetachPortType = 0;

//Spin Missile By This RPM After Launch Clearance
int spinAmount = 0;

//Whether to perform a vertical takeoff for the launching procedure
bool verticalTakeoff = false;

//Whether to reduce the GFD beam width by the radius of your missile
bool excludeGFDMissileRadius = false;

//Whether to fly straight first until LOCK command is given via R_TARGET Text Panel to start homing. Activating this will ignore launchSeconds variable
bool waitForHomingTrigger = false;

//Whether to allow missile to read Private Title of the R_TARGET to get command during missile homing
bool enableMissileCommand = false;

//Whether to average out the target direction when calculating lead to mitigate impulse jammers
bool antiJamUseAverage = true;

//------------------------------ Reference Block Name Configuration ------------------------------

string strShipRefRemoteControl = "R_REMOTE";
string strShipRefTargetPanel = "R_TARGET";
string strShipRefCamera = "R_CAMERA";

string strShipRefNotMissile = "NOT_MISSILE";

//This line configures missile trigger commands to allow missile to perform action based on proximity and duration
string missileTriggerCommands = "";

//By default all gyroscopes, thrusters and merge blocks will be considered for use. Setting a value here limits the script to use specific set of blocks
const string strGyroscopesPrefix = "";
const string strThrustersPrefix = "";
const string strDetachPort = "";
const string strDirectionRefBlock = "";
const string strLockOnSensor = "R_SENSOR";

//For missile lock alert
const string strLockTriggerBlock = "R_ALERT";
const string strLockTriggerAction = "PlaySound";

//For debugging purposes
const string strStatusDisplayPrefix = "<D>";

//------------------------------ Missile Handling Configuration ------------------------------

double driftVectorReduction = 1.2;
double launchSeconds = 1;

//By default, script will determine whether to use these parameters. Setting a value forces script to use it
bool? boolDrift = null;
bool? boolLeadTarget = null;
bool? boolNaturalDampener = null;

//------------------------------ GetFreeDestination Lock On Configuration ------------------------------

const float GFD_MIN_LOCK_DISTANCE = 100;
const float GFD_MAX_LOCK_DISTANCE = 20000;
float GFD_BEAM_WIDTH = 0f;

const int GFD_STABILIZATION_TICKS = 30;
const float GFD_STABILIZATION_DISTANCE = 1;

//The number of ticks to run for every GFD lock-on algorithm used (For full lock-on only, initial lock-on not applicable). Example, 60 means the GFD will be called once every 60 ticks
int GFD_LOCK_ON_INTERVAL = 15;

//------------------------------ Above Is User Configuration Section. This Section Is For PID Tuning ------------------------------

const double DEF_SMALL_GRID_P = 300;
const double DEF_SMALL_GRID_I = 0.1;
const double DEF_SMALL_GRID_D = 100;

const double DEF_BIG_GRID_P = 50;
const double DEF_BIG_GRID_I = 0.5;
const double DEF_BIG_GRID_D = 4;

bool useDefaultPIDValues = true;

double AIM_P = 0;
double AIM_I = 0;
double AIM_D = 0;
double AIM_LIMIT = 60;

double INTEGRAL_WINDUP_LIMIT = 0;

//------------------------------ Script Parameters Configuration ------------------------------

const int MERGE_SEPARATE_WAIT_THRESHOLD = 60;
const int ANTI_JAM_AVERAGE_ARRAY_SIZE = 6;

bool outputMissileStatus = false;

//------------------------------ Below Is Main Script Body ------------------------------

IMyRemoteControl shipRefRemoteControl = null;
IMyTextPanel shipRefTargetPanel = null;
IMyTerminalBlock shipRefCamera = null;

IMyRemoteControl remoteControl = null;
IMyTerminalBlock refForwardBlock = null;
IMyTerminalBlock refDownwardBlock = null;
IMySensorBlock lockOnSensor = null;

IMyTerminalBlock alertBlock = null;
IMyTerminalBlock statusDisplay = null;

IMyTerminalBlock notMissile = null;
double notMissileRadius = 0;

List<IMyTerminalBlock> gyroscopes = null;
string[] gyroYawField = null;
string[] gyroPitchField = null;
string[] gyroRollField = null;
float[] gyroYawFactor = null;
float[] gyroPitchFactor = null;
float[] gyroRollFactor = null;

List<IMyTerminalBlock> thrusters = null;
float[] thrustValues = null;

List<IMyTerminalBlock> launchThrusters = null;

MatrixD refWorldMatrix = default(MatrixD);
MatrixD refLookAtMatrix = default(MatrixD);
bool refForwardReverse = false;

bool cameraIsTurret = false;

Vector3D naturalGravity = new Vector3D();
double naturalGravityLength = 0;

Vector3D driftVector = new Vector3D();
double speed = 0;
double rpm = 0;

Vector3D lastNormal = new Vector3D();

IMyCubeGrid lockOnGrid = null;
IMyCubeGrid lastDetectedGrid = null;

Vector3D remoteTargetLocation = new Vector3D();
Vector3D remoteTargetPosition = new Vector3D();
double remoteTargetRadius = 0;
bool remoteTargetFound = false;

Vector3D lastGFDPosition = new Vector3D();
Vector3D currGFDPosition = new Vector3D();
int lastGFDPositionClock = 0;
int currGFDPositionClock = 0;

float adjustedBeamWidth = 25f;

Vector3D targetPosition = new Vector3D();
Vector3D lastTargetPosition = new Vector3D();

bool targetPositionSet = false;
int lastTargetPositionClock = 0;

Vector3D targetVector = new Vector3D();
double distToTarget = 0;

Vector3D targetDirection = new Vector3D();
double targetSpeed = 0;

int avgTargetDirectionIndex = 0;
Vector3D[] avgTargetDirectionArr = null;
Vector3D avgTargetDirectionSum = new Vector3D();

double targetYawAngle = 0;
double targetPitchAngle = 0;
double targetRollAngle = 0;

double lastYawError = 0;
double lastYawIntegral = 0;
double lastPitchError = 0;
double lastPitchIntegral = 0;
double lastRollError = 0;
double lastRollIntegral = 0;

int subCounter = 0;
int subMode = 0;
int mode = 0;
int clock = 0;
bool init = false;

IMyTerminalBlock detachBlockLocal = null;
IMyTerminalBlock detachBlockRemote = null;

List<KeyValuePair<double, string[]>> rpmTriggerList = null;
List<KeyValuePair<double, string[]>> distTriggerList = null;
List<KeyValuePair<int, string[]>> timeTriggerList = null;
bool haveTriggerCommands = false;

string nameMatcher = "";

Random rnd = new Random();

const double RPM_FACTOR = 1800 / Math.PI;
const double ACOS_FACTOR = 180 / Math.PI;
const float GYRO_FACTOR = (float)(Math.PI / 30);

Vector3D Y_VECTOR = new Vector3D(0, -1, 0);
Vector3D Z_VECTOR = new Vector3D(0, 0, -1);
Vector3D POINT_ZERO = new Vector3D(0, 0, 0);

void Main(string arguments)
{
//---------- Initialization And General Controls ----------

if (!init)
{
if (subMode == 0)  //Check for configuration command
{
subMode = 1;

if (arguments != null && arguments.Length > 0)
{
ProcessConfigurationCommand(arguments);
return;
}
}

if (subMode == 1)  //Missile still on launching ship's grid
{
if (!InitLaunchingShipRefBlocks())
{
throw new Exception("--- Initialization Failed ---");
}

if (!DetachFromGrid())
{
throw new Exception("--- Initialization Failed ---");
}

subCounter = 0;
subMode = (missileDetachPortType == 99 ? 3 : 2);
return;
}
else if (subMode == 2)  //Missile waiting for successful detachment from launching ship
{
bool isDetached = false;

if (notMissile != null)
{
isDetached = (notMissile.CubeGrid != Me.CubeGrid);
}
else
{
isDetached = (detachBlockLocal == null || detachBlockRemote == null || detachBlockLocal.CubeGrid != detachBlockRemote.CubeGrid);
}

if (isDetached)
{
subMode = 3;
return;
}
else
{
subCounter++;

if (subCounter >= MERGE_SEPARATE_WAIT_THRESHOLD)
{
Echo("Error: Missile detach failed.");
throw new Exception("--- Initialization Failed ---");
}

return;
}
}
else if (subMode == 3)  //Missile successfully detached and currently initializing
{
if (missileDetachPortType == 3 || missileDetachPortType == 4)
{
DetachLockedConnectors();
}

if (notMissile != null)
{
notMissileRadius = ComputeBlockGridDiagonalVector(notMissile).Length() / 2.0;
}

if (!InitMissileBlocks())
{
throw new Exception("--- Initialization Failed ---");
}
}

if (missileLaunchType == 99)
{
subMode = 0;
mode = 99;
clock = 0;
}
else
{
if (excludeGFDMissileRadius)
{
GFD_BEAM_WIDTH = (float)(ComputeBlockGridDiagonalVector(Me).Length() / -2);
}

if (antiJamUseAverage)
{
avgTargetDirectionArr = new Vector3D[ANTI_JAM_AVERAGE_ARRAY_SIZE];
avgTargetDirectionIndex = 0;
}

subCounter = (waitForHomingTrigger ? int.MaxValue : (int)(launchSeconds * 60));
FireThrusters(verticalTakeoff ? launchThrusters : thrusters, true);

subMode = (missileLaunchType == 0 || missileLaunchType == 2 ? 1 : 0);
mode = 1;
clock = 0;
}

init = true;
return;
}

//---------- Modes And Controls ----------

clock += 1;

CalculateParameters();

if (enableMissileCommand)
{
ProcessMissileCommand(shipRefTargetPanel.GetPrivateTitle());
}

if (mode == 1)  //Launching
{
if (waitForHomingTrigger)
{
if ("LOCK".Equals(shipRefTargetPanel.GetPublicTitle()) || "LOCK".Equals(shipRefTargetPanel.GetPrivateTitle()))
{
subCounter = 0;
}
}

if (subMode == 1)
{
if (subCounter > 0)
{
subCounter--;
}
else if (targetPositionSet)
{
if (verticalTakeoff)
{
FireThrusters(launchThrusters, false);
FireThrusters(thrusters, true);
}

SetGyroOverride(true);

if (spinAmount > 0)
{
SetGyroRoll(spinAmount);
}

if (missileLaunchType == 2)
{
TriggerLockAlert();

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}
}

subCounter = 0;
subMode = 0;
mode = 10 + missileLaunchType;
}
else
{
remoteTargetLocation = shipRefRemoteControl.WorldMatrix.Forward + shipRefRemoteControl.WorldMatrix.Translation;

if (FindGFDTarget(shipRefRemoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH))
{
targetPositionSet = true;
}
else
{
targetPositionSet = false;
}

if (targetPositionSet)
{
if (notMissile != null && (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() <= notMissileRadius)
{
targetPositionSet = false;
}
else if ((remoteTargetPosition - shipRefRemoteControl.WorldMatrix.Translation).Length() < GFD_MIN_LOCK_DISTANCE || (GetMissileMidPoint() - remoteTargetPosition).Length() < 1)
{
targetPositionSet = false;
}
}

if (targetPositionSet && missileLaunchType == 2)
{
Vector3D checkTargetPosition = remoteTargetPosition;

if (FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, remoteTargetRadius, out remoteTargetPosition, out remoteTargetLocation, GFD_BEAM_WIDTH))
{
if ((remoteTargetPosition - checkTargetPosition).Length() >= GFD_STABILIZATION_DISTANCE)
{
targetPositionSet = false;
}
}
else
{
targetPositionSet = false;
}
}

if (targetPositionSet)
{
targetPosition = remoteTargetPosition;
}

if (verticalTakeoff && subCounter == 0)
{
FireThrusters(launchThrusters, false);
FireThrusters(thrusters, true);

subCounter = -1;
}

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}
}
}
else
{
if (subCounter > 0)
{
subCounter--;
}
else
{
if (verticalTakeoff)
{
FireThrusters(launchThrusters, false);
FireThrusters(thrusters, true);
}

SetGyroOverride(true);

if (spinAmount > 0)
{
SetGyroRoll(spinAmount);
}

if (missileLaunchType == 3 || missileLaunchType == 4 || missileLaunchType == 80)
{
if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = 1000000000;
ExecuteTriggerCommand(missileTriggerCommands);
}
}

lastTargetPosition = targetPosition = GetFlyStraightVector();

subCounter = 0;
subMode = 0;
mode = 10 + missileLaunchType;
}
}
}
else if (mode == 10)  //Active Radar Homing With Initial Shipborne GFD Lock-On
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

if (subMode == 0)  //Initial Lock-On
{
if (!targetPositionSet)
{
remoteTargetLocation = shipRefRemoteControl.WorldMatrix.Forward + shipRefRemoteControl.WorldMatrix.Translation;
}

if (FindGFDTarget(shipRefRemoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH))
{
targetPositionSet = true;
}
else
{
targetPositionSet = false;
}

if (targetPositionSet)
{
if (notMissile != null && (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() <= notMissileRadius)
{
targetPositionSet = false;
}
else if ((remoteTargetPosition - shipRefRemoteControl.WorldMatrix.Translation).Length() < GFD_MIN_LOCK_DISTANCE || (GetMissileMidPoint() - remoteTargetPosition).Length() < 1)
{
targetPositionSet = false;
}
}

bool missileLocked = false;

if (targetPositionSet)
{
targetPosition = remoteTargetPosition;

if (FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, remoteTargetRadius, out remoteTargetPosition, out remoteTargetLocation, GFD_BEAM_WIDTH))
{
if ((remoteTargetPosition - targetPosition).Length() < GFD_STABILIZATION_DISTANCE)
{
missileLocked = true;
}
}
}

if (missileLocked)
{
subCounter += 1;

if (subCounter >= GFD_STABILIZATION_TICKS)
{
TriggerLockAlert();

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}

lastTargetPosition = lastGFDPosition = targetPosition;
lastTargetPositionClock = lastGFDPositionClock = clock;

subCounter = 0;
subMode = 1;
}
}
else
{
subCounter = 0;
}

if (targetPositionSet)
{
CalculateTargetParameters();
targetPositionSet = true;
}
else
{
CalculateTargetParameters();
}
}
else if (subMode == 1)  //Active Radar Homing
{
PerformARHLogic();

CalculateTargetParameters();
}

AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 11)  //Active Radar Homing With Initial Shipborne GPS Coordinates Lock-On
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

if (subMode == 0)  //Initial Lock-On
{
Vector3D parsedVector;
if (ParseCoordinates(shipRefTargetPanel.GetPublicTitle(), out parsedVector))
{
targetPosition = parsedVector;
targetPositionSet = true;
}
else
{
lastTargetPosition = targetPosition = GetFlyStraightVector();
targetPositionSet = false;
}

if (targetPositionSet && (GetMissileMidPoint() - targetPosition).Length() < 1)
{
lastTargetPosition = targetPosition = GetFlyStraightVector();
targetPositionSet = false;
}

bool missileLocked = false;

if (targetPositionSet)
{
if (FindGFDTarget(remoteControl, targetPosition, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH))
{
if ((remoteTargetPosition - targetPosition).Length() < remoteTargetRadius)
{
missileLocked = true;
}
}
}

if (missileLocked)
{
subCounter += 1;

if (subCounter >= GFD_STABILIZATION_TICKS)
{
TriggerLockAlert();

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}

lastTargetPosition = lastGFDPosition = targetPosition = remoteTargetPosition;
lastTargetPositionClock = lastGFDPositionClock = clock;

subCounter = 0;
subMode = 1;
}
}
else
{
subCounter = 0;
}
}
else if (subMode == 1)  //Active Radar Homing
{
PerformARHLogic();
}

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 12)  //Semi-Active Radar Homing Using Shipborne GFD Lock-On
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

if (GFD_LOCK_ON_INTERVAL > 1)
{
if (subCounter <= 1)
{
remoteTargetFound = FindGFDTarget(shipRefRemoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH);

if (remoteTargetFound)
{
if (clock - lastGFDPositionClock > GFD_LOCK_ON_INTERVAL)
{
lastGFDPosition = remoteTargetPosition;
lastGFDPositionClock = clock;
}
else
{
currGFDPosition = remoteTargetPosition;
currGFDPositionClock = clock;

subCounter = GFD_LOCK_ON_INTERVAL;
}
}
}
else
{
Vector3D direction = (currGFDPosition - lastGFDPosition) / (currGFDPositionClock - lastGFDPositionClock);
remoteTargetPosition += direction;
remoteTargetLocation += direction;

subCounter--;
}
}
else
{
remoteTargetFound = FindGFDTarget(shipRefRemoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH);
}

if (notMissile != null && (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() <= notMissileRadius)
{
remoteTargetFound = false;
}
else if ((GetMissileMidPoint() - remoteTargetPosition).Length() < 1)
{
remoteTargetFound = false;
}

if (remoteTargetFound)
{
targetPosition = remoteTargetPosition;
targetPositionSet = true;
}
else
{
targetPositionSet = false;
subCounter = 0;
}

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 13)  //Camera Guided Mode
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

Vector3D cameraPosition = shipRefCamera.WorldMatrix.Translation;
Vector3D cameraForward = (cameraIsTurret ? CalculateTurretViewVector(shipRefCamera as IMyLargeTurretBase) : shipRefCamera.WorldMatrix.Forward);

Vector3D shipToMissileVector = remoteControl.GetPosition() - cameraPosition;
Vector3D missileToViewLineVector = Vector3D.Reject(shipToMissileVector, cameraForward);

double extraDistanceExtend = Math.Max(5.6713 * missileToViewLineVector.Length(), speed * 2);
extraDistanceExtend += (shipToMissileVector - missileToViewLineVector).Length();

targetPosition = cameraPosition + (cameraForward * extraDistanceExtend);
targetPositionSet = true;

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 14)  //Cruise Mode
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

Vector3D parsedVector;
if (ParseCoordinates(shipRefTargetPanel.GetPublicTitle(), out parsedVector))
{
targetPosition = parsedVector;
targetPositionSet = true;
}

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 15)  //Active Radar Homing With Initial Shipborne Camera Guidance
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

if (subMode == 0)  //Initial Camera Guidance
{
Vector3D cameraPosition = shipRefCamera.WorldMatrix.Translation;
Vector3D cameraForward = (cameraIsTurret ? CalculateTurretViewVector(shipRefCamera as IMyLargeTurretBase) : shipRefCamera.WorldMatrix.Forward);

Vector3D shipToMissileVector = remoteControl.GetPosition() - cameraPosition;
Vector3D viewLineToMissileVector = Vector3D.Reject(shipToMissileVector, cameraForward);

double extraDistanceExtend = Math.Max(5.6713 * viewLineToMissileVector.Length(), speed * 2);
extraDistanceExtend += (shipToMissileVector - viewLineToMissileVector).Length();

targetPosition = cameraPosition + (cameraForward * extraDistanceExtend);
targetPositionSet = true;

Vector3D cameraSafePoint = cameraPosition + (cameraForward * GFD_MIN_LOCK_DISTANCE);
Vector3D safePointToMissile = remoteControl.GetPosition() - cameraSafePoint;

if (safePointToMissile.Dot(cameraForward) > 0)
{
remoteTargetLocation = (cameraForward * subCounter * 100) + cameraSafePoint;

if (FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, adjustedBeamWidth))
{
if (notMissile == null || (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() > notMissileRadius)
{
if (cameraForward.Dot(remoteTargetPosition - cameraSafePoint) > 0)
{
if (new RayD(cameraSafePoint, cameraForward).Intersects(new BoundingSphereD(remoteTargetPosition, remoteTargetRadius)) != null)
{
TriggerLockAlert();

targetPosition = remoteTargetPosition;

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}

lastTargetPosition = lastGFDPosition = targetPosition;
lastTargetPositionClock = lastGFDPositionClock = clock;

subCounter = 0;
subMode = 1;
}
else
{
adjustedBeamWidth = 0f;
}
}
}
}
else if (remoteTargetRadius == -1)
{
adjustedBeamWidth = 0f;
}

if (subMode == 0 && (subCounter % 6 == 0))
{
remoteTargetLocation = (cameraForward * (((subCounter / 6) * 1000) + 7000)) + cameraSafePoint;

if (FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, adjustedBeamWidth))
{
if (notMissile == null || (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() > notMissileRadius)
{
if (cameraForward.Dot(remoteTargetPosition - cameraSafePoint) > 0)
{
if (new RayD(cameraSafePoint, cameraForward).Intersects(new BoundingSphereD(remoteTargetPosition, remoteTargetRadius)) != null)
{
TriggerLockAlert();

targetPosition = remoteTargetPosition;

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}

lastTargetPosition = lastGFDPosition = targetPosition;
lastTargetPositionClock = lastGFDPositionClock = clock;

subCounter = 0;
subMode = 1;
}
}
}
}
}

subCounter++;
if (subCounter >= 60)
{
adjustedBeamWidth = 25f;
subCounter = 0;
}
}
}
else if (subMode == 1)  //Active Radar Homing
{
PerformARHLogic();
}

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 16)  //Active Radar Homing With Initial Shipborne GFD Lock-On And Camera Guidance
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

if (subMode == 0)  //Initial Lock-On And Camera Guidance
{
Vector3D cameraPosition = shipRefCamera.WorldMatrix.Translation;
Vector3D cameraForward = (cameraIsTurret ? CalculateTurretViewVector(shipRefCamera as IMyLargeTurretBase) : shipRefCamera.WorldMatrix.Forward);

Vector3D shipToMissileVector = remoteControl.GetPosition() - cameraPosition;
Vector3D viewLineToMissileVector = Vector3D.Reject(shipToMissileVector, cameraForward);

double extraDistanceExtend = Math.Max(5.6713 * viewLineToMissileVector.Length(), speed * 2);
extraDistanceExtend += (shipToMissileVector - viewLineToMissileVector).Length();

targetPosition = cameraPosition + (cameraForward * extraDistanceExtend);
targetPositionSet = true;

remoteTargetLocation = targetPosition;

float initialLockBeamWidth = (excludeGFDMissileRadius ? GFD_BEAM_WIDTH : (float)(ComputeBlockGridDiagonalVector(Me).Length() / -2));
bool missileLocked = false;

if (FindGFDTarget(shipRefRemoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, initialLockBeamWidth))
{
if (notMissile == null || (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() > notMissileRadius)
{
if ((remoteTargetPosition - shipRefRemoteControl.WorldMatrix.Translation).Length() >= GFD_MIN_LOCK_DISTANCE && (GetMissileMidPoint() - remoteTargetPosition).Length() >= 1)
{
Vector3D shipLockTargetPosition = remoteTargetPosition;

if (FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, remoteTargetRadius, out remoteTargetPosition, out remoteTargetLocation, initialLockBeamWidth))
{
if ((remoteTargetPosition - shipLockTargetPosition).Length() < GFD_STABILIZATION_DISTANCE)
{
missileLocked = true;
}
}
}
}
}

if (missileLocked)
{
subCounter += 1;

if (subCounter >= GFD_STABILIZATION_TICKS)
{
TriggerLockAlert();

targetPosition = remoteTargetPosition;

if (missileTriggerCommands != null && missileTriggerCommands.Length > 0)
{
distToTarget = (refWorldMatrix.Translation - targetPosition).Length();
ExecuteTriggerCommand(missileTriggerCommands);
}

lastTargetPosition = lastGFDPosition = remoteTargetPosition;
lastTargetPositionClock = lastGFDPositionClock = clock;

subCounter = 0;
subMode = 1;
}
}
else
{
subCounter = 0;
}

if (targetPositionSet)
{
CalculateTargetParameters();
targetPositionSet = true;
}
else
{
CalculateTargetParameters();
}
}
else if (subMode == 1)  //Active Radar Homing
{
PerformARHLogic();

CalculateTargetParameters();
}

AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 20)  //Sensor Chase
{
targetPosition = GetLockOnGridPosition();
targetPositionSet = true;

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}
else if (mode == 90)  //Evil NPC Mode
{
if (lockOnSensor != null && CheckLockOnSensor())
{
mode = 20;
}

targetPositionSet = remoteControl.GetNearestPlayer(out targetPosition);

CalculateTargetParameters();
AimAtTarget();

if (boolNaturalDampener == true)
{
AimAtNaturalGravity();
}

if (haveTriggerCommands)
{
ProcessTriggerCommands();
}
}

if (statusDisplay != null)
{
if (mode == 0)
{
DisplayStatus("Idle");
}
else if (mode == 1)
{
DisplayStatus("Launching");
}
else if (mode == 10 || mode == 11 || mode == 16)
{
if (subMode == 0)
{
DisplayStatus("Initial Lock - Stable Ticks: " + subCounter);
}
else if (subMode == 1)
{
DisplayStatus((remoteTargetFound ? "Lock" : "Trace") + ": [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else
{
DisplayStatus("-");
}
}
else if (mode == 12)
{
DisplayStatus((remoteTargetFound ? "Lock" : "Trace") + ": [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else if (mode == 13)
{
DisplayStatus("Camera");
}
else if (mode == 14)
{
DisplayStatus("Aim: [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else if (mode == 15)
{
if (subMode == 0)
{
DisplayStatus("Initial Camera Lock");
}
else if (subMode == 1)
{
DisplayStatus((remoteTargetFound ? "Lock" : "Trace") + ": [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else
{
DisplayStatus("-");
}
}
else if (mode == 20)
{
DisplayStatus("Sensor: [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else if (mode == 90)
{
DisplayStatus("Aim: [" + Math.Round(targetPosition.GetDim(0), 2) + "," + Math.Round(targetPosition.GetDim(1), 2) + "," + Math.Round(targetPosition.GetDim(2), 2) + "]");
}
else
{
DisplayStatus("-");
}
}

if (outputMissileStatus)
{
string statusCode;
switch (mode)
{
case 0:
statusCode = "-";
break;
case 1:
if (waitForHomingTrigger)
{
statusCode = "W";
}
else if (subCounter > 0)
{
statusCode = "F";
}
else
{
statusCode = "K";
}
break;
case 10:
case 11:
if (subMode == 0 || remoteTargetFound)
{
statusCode = "L";
}
else
{
statusCode = "T";
}
break;
case 12:
if (remoteTargetFound)
{
statusCode = "L";
}
else
{
statusCode = "T";
}
break;
case 13:
case 14:
statusCode = "C";
break;
case 15:
if (subMode == 0)
{
statusCode = "C";
}
else if (remoteTargetFound)
{
statusCode = "L";
}
else
{
statusCode = "T";
}
break;
case 16:
if (subMode == 0)
{
statusCode = "C";
}
else if (remoteTargetFound)
{
statusCode = "L";
}
else
{
statusCode = "T";
}
break;
case 20:
statusCode = "S";
break;
case 90:
statusCode = "L";
break;
default:
statusCode = "-";
break;
}
Echo("ST:" + mode + ":" + subMode + ":" + (waitForHomingTrigger ? 0 : subCounter) + ":" + clock + ":" + statusCode + ":" +
Math.Round(targetPosition.GetDim(0), 5) + ":" + Math.Round(targetPosition.GetDim(1), 5) + ":" + Math.Round(targetPosition.GetDim(2), 5) + ":" +
Math.Round(remoteTargetRadius, 5) + ":");
}
}

//------------------------------ Miscellaneous Methods ------------------------------

void DisplayStatus(string statusMsg)
{
if (statusDisplay != null)
{
statusDisplay.SetCustomName(strStatusDisplayPrefix + " Mode: " + mode + ", " + statusMsg);
}
}

bool CheckLockOnSensor()
{
bool switchToGrid = false;
#if STABLE
IMyCubeGrid detectedGrid = lockOnSensor.LastDetectedEntity as IMyCubeGrid;
#else
IMyCubeGrid detectedGrid = null;
#endif

if (lastDetectedGrid != detectedGrid)
{
if (detectedGrid != null)
{
if (lockOnGrid == null)
{
switchToGrid = true;
}
else
{
double detectedGridLength = (detectedGrid.GridIntegerToWorld(detectedGrid.Min) - detectedGrid.GridIntegerToWorld(detectedGrid.Max)).Length();
double lockOnGridLength = (lockOnGrid.GridIntegerToWorld(lockOnGrid.Min) - lockOnGrid.GridIntegerToWorld(lockOnGrid.Max)).Length();

switchToGrid = (detectedGridLength > lockOnGridLength);
}

if (switchToGrid)
{
lockOnGrid = detectedGrid;
return true;
}
}
lastDetectedGrid = detectedGrid;
}

return false;
}

void PerformARHLogic()
{
if (GFD_LOCK_ON_INTERVAL > 1)
{
if (subCounter <= 1)
{
remoteTargetFound = FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH);

if (notMissile != null && (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() <= notMissileRadius)
{
remoteTargetFound = false;
}

if (remoteTargetFound)
{
if (clock - lastGFDPositionClock > GFD_LOCK_ON_INTERVAL)
{
lastGFDPosition = remoteTargetPosition;
lastGFDPositionClock = clock;
}
else
{
currGFDPosition = remoteTargetPosition;
currGFDPositionClock = clock;

subCounter = GFD_LOCK_ON_INTERVAL;
}
}
}
else
{
Vector3D direction = (currGFDPosition - lastGFDPosition) / (currGFDPositionClock - lastGFDPositionClock);
remoteTargetPosition += direction;
remoteTargetLocation += direction;

subCounter--;
}
}
else
{
remoteTargetFound = FindGFDTarget(remoteControl, remoteTargetLocation, GFD_MAX_LOCK_DISTANCE, out remoteTargetPosition, out remoteTargetRadius, out remoteTargetLocation, GFD_BEAM_WIDTH);

if (notMissile != null && (remoteTargetPosition - ComputeBlockGridMidPoint(notMissile)).Length() <= notMissileRadius)
{
remoteTargetFound = false;
}
}

if (remoteTargetFound)
{
targetPosition = remoteTargetPosition;
targetPositionSet = true;
}
else
{
targetPositionSet = false;
subCounter = 0;
}
}

void TriggerLockAlert()
{
if (alertBlock != null)
{
if (alertBlock.HasAction(strLockTriggerAction))
{
alertBlock.ApplyAction(strLockTriggerAction);
}
}
}

Vector3D GetLockOnGridPosition()
{
return lockOnGrid.GridIntegerToWorld((lockOnGrid.Min + lockOnGrid.Max) / 2);
}

Vector3D GetMissileMidPoint()
{
return (Me.CubeGrid.GridIntegerToWorld(Me.CubeGrid.Min) + Me.CubeGrid.GridIntegerToWorld(Me.CubeGrid.Max)) / 2;
}

Vector3D GetFlyStraightVector()
{
return (driftVector * 1000) + remoteControl.GetPosition();
}

//------------------------------ Missile And Target Information Methods ------------------------------

void CalculateParameters()
{
//---------- Calculate Missile Related Variables ----------

refWorldMatrix = refForwardBlock.WorldMatrix;
refLookAtMatrix = MatrixD.CreateLookAt(POINT_ZERO, (refForwardReverse ? refWorldMatrix.Backward : refWorldMatrix.Forward), refWorldMatrix.Up);

driftVector = remoteControl.GetShipVelocities().LinearVelocity;
speed = driftVector.Length();

rpm = Math.Acos(lastNormal.Dot(refWorldMatrix.Up)) * RPM_FACTOR;
lastNormal = refWorldMatrix.Up;

naturalGravity = remoteControl.GetNaturalGravity();
naturalGravityLength = naturalGravity.Length();
naturalGravity = (naturalGravityLength > 0 ? naturalGravity / naturalGravityLength : POINT_ZERO);
}

void CalculateTargetParameters()
{
//---------- Calculate Target Parameters ----------

if (boolLeadTarget == true)
{
if (targetPositionSet)
{
targetDirection = targetPosition - lastTargetPosition;
if (antiJamUseAverage)
{
avgTargetDirectionSum -= avgTargetDirectionArr[avgTargetDirectionIndex];
avgTargetDirectionArr[avgTargetDirectionIndex] = targetDirection;
avgTargetDirectionSum += avgTargetDirectionArr[avgTargetDirectionIndex];
avgTargetDirectionIndex++;
if (avgTargetDirectionIndex >= avgTargetDirectionArr.Length)
{
avgTargetDirectionIndex = 0;
}
targetDirection = avgTargetDirectionSum / avgTargetDirectionArr.Length;
}
targetSpeed = targetDirection.Length();

if (targetSpeed > 0)
{
targetDirection = targetDirection / targetSpeed;
targetSpeed = targetSpeed * 60 / (clock - lastTargetPositionClock);
}

lastTargetPosition = targetPosition;
lastTargetPositionClock = clock;

targetPositionSet = false;
}
else
{
targetPosition = lastTargetPosition + ((targetDirection * targetSpeed) / 60 * (clock - lastTargetPositionClock));
}

if (targetSpeed > 0)
{
Vector3D aimPosition = ComputeIntersectionPoint(targetDirection, targetPosition, targetSpeed, remoteControl.GetPosition(), speed);
if (!Double.IsNaN(aimPosition.Sum))
{
targetPosition = aimPosition;
}
}
}

targetVector = targetPosition - remoteControl.GetPosition();
distToTarget = targetVector.Length();
targetVector = targetVector / distToTarget;

if (boolDrift == true && speed >= 5)
{
targetVector = (targetVector * speed) - (driftVector / driftVectorReduction);
targetVector.Normalize();
}

targetVector = Vector3D.TransformNormal(targetVector, refLookAtMatrix);
targetVector.Normalize();

if (Double.IsNaN(targetVector.Sum))
{
targetVector = new Vector3D(Z_VECTOR);
}
}

Vector3D CalculateTurretViewVector(IMyLargeTurretBase turret)
{
Vector3D direction;
Vector3D.CreateFromAzimuthAndElevation(turret.Azimuth, turret.Elevation, out direction);

return Vector3D.TransformNormal(direction, turret.WorldMatrix);
}

//------------------------------ Missile Lock-On And Leading Methods ------------------------------

//Two beam GFD lock-on method. This is the optimized GFD lock-on algorithm.
bool FindGFDTarget(IMyRemoteControl remoteControl, Vector3D targetLocation, float targetDistance, out Vector3D targetPosition, out double targetRadius, out Vector3D nextTargetLocation, float beamWidth = 0f)
{
Vector3D remotePosition = remoteControl.WorldMatrix.Translation;
Vector3D targetVector1 = targetLocation - remotePosition;
targetVector1.Normalize();

Vector3D targetPoint1 = (targetVector1 * targetDistance) + remotePosition;
#if STABLE
Vector3D offsetTarget1 = remoteControl.GetFreeDestination(targetPoint1, targetDistance, beamWidth);
#else
Vector3D offsetTarget1 = targetPoint1;
#endif
double remoteToOffsetDist1 = (offsetTarget1 - remotePosition).Length();

if ((targetPoint1 - offsetTarget1).Length() < 1)
{
targetPosition = targetLocation;
targetRadius = 0;

nextTargetLocation = targetLocation;

return false;
}

Vector3D estimatedPoint1 = (targetVector1 * remoteToOffsetDist1) + remotePosition;
Vector3D targetNormal1 = targetVector1.Cross(offsetTarget1 - estimatedPoint1);
targetNormal1.Normalize();

Vector3D targetVector2 = offsetTarget1 - estimatedPoint1;
targetVector2 = targetVector2 / targetVector2.Length();
targetVector2 = (estimatedPoint1 - targetVector2 + targetNormal1) - remotePosition;
targetVector2.Normalize();

Vector3D targetPoint2 = (targetVector2 * targetDistance) + remotePosition;
#if STABLE
Vector3D offsetTarget2 = remoteControl.GetFreeDestination(targetPoint2, targetDistance, beamWidth);
#else
Vector3D offsetTarget2 = targetPoint2;
#endif
double remoteToOffsetDist2 = (offsetTarget2 - remotePosition).Length();

if ((targetPoint2 - offsetTarget2).Length() < 1)
{
targetPosition = targetLocation;
targetRadius = 0;

nextTargetLocation = targetLocation;

return false;
}

if (Math.Abs(remoteToOffsetDist1 - remoteToOffsetDist2) < 0.000001)
{
targetPosition = targetLocation;
targetRadius = -1;

nextTargetLocation = targetLocation;

return false;
}

Vector3D estimatedPoint2 = (targetVector2 * remoteToOffsetDist2) + remotePosition;
Vector3D targetNormal2 = targetVector2.Cross(offsetTarget2 - estimatedPoint2);
targetNormal2.Normalize();

Vector3D intersectVector = targetNormal1.Cross(targetNormal2);

double zTopValue = (remotePosition - offsetTarget2).LengthSquared() - (remotePosition - offsetTarget1).LengthSquared();
double zBottomValue = 2 * intersectVector.Dot(offsetTarget2 - offsetTarget1);

if (zBottomValue != 0)
{
targetPoint1 = remotePosition + (intersectVector * (zTopValue / zBottomValue));
if (Double.IsNaN(targetPoint1.Sum))
{
targetPosition = targetLocation;
targetRadius = 0;

nextTargetLocation = targetLocation;

return false;
}
else
{
double b = targetVector1.Dot(2 * remotePosition - targetPoint1 - offsetTarget1);
double c = (remotePosition - targetPoint1).Dot(remotePosition - offsetTarget1);
double u = Math.Sqrt((b * b) - (4 * c));

targetPosition = targetPoint1;
targetRadius = ((Math.Min(-b + u, -b - u) * 0.5 * targetVector1) + remotePosition - targetPoint1).Length() - beamWidth;

nextTargetLocation = targetPoint1 + targetNormal1;

return true;
}
}
else
{
targetPosition = targetLocation;
targetRadius = 0;

nextTargetLocation = targetLocation;

return false;
}
}

//One beam GFD lock-on shortcut method. This is the shortcut GFD lock-on where it uses a pre-calculated radius.
bool FindGFDTarget(IMyRemoteControl remoteControl, Vector3D targetLocation, float targetDistance, double targetRadius, out Vector3D targetPosition, out Vector3D nextTargetLocation, float beamWidth = 0f)
{
Vector3D remotePosition = remoteControl.WorldMatrix.Translation;
Vector3D targetVector1 = targetLocation - remotePosition;
targetVector1.Normalize();

Vector3D targetPoint1 = (targetVector1 * targetDistance) + remotePosition;
#if STABLE
Vector3D offsetTarget1 = remoteControl.GetFreeDestination(targetPoint1, targetDistance, beamWidth);
#else
Vector3D offsetTarget1 = targetPoint1;
#endif

if ((targetPoint1 - offsetTarget1).Length() < 1)
{
targetPosition = targetLocation;

nextTargetLocation = targetLocation;

return false;
}

double b = 2 * targetVector1.Dot(remotePosition - offsetTarget1);
double c = (remotePosition - offsetTarget1).LengthSquared() - Math.Pow(Math.Max((targetRadius + beamWidth) * 0.5, 20), 2);
double u = Math.Sqrt(Math.Round((b * b) - (4 * c), 4));
Vector3D intersectPoint = (Math.Min(-b + u, -b - u) * 0.5 * targetVector1) + remotePosition;

Vector3D offsetVector1 = offsetTarget1 - intersectPoint;
offsetVector1.Normalize();

Vector3D coordVector = Vector3D.Reject(targetVector1, offsetVector1);
coordVector.Normalize();

targetPosition = intersectPoint + (coordVector * (targetRadius + beamWidth));

if (Double.IsNaN(targetPosition.Sum))
{
targetPosition = targetLocation;
nextTargetLocation = targetLocation;

return false;
}
else
{
nextTargetLocation = targetPosition + coordVector;

return true;
}
}

Vector3D ComputeIntersectionPoint(Vector3D targetDirection, Vector3D targetLocation, double targetSpeed, Vector3D currentLocation, double currentSpeed)
{
//---------- Calculate Impact Point ----------

//targetDirection Must Be Normalized
double a = (targetSpeed * targetSpeed) - (currentSpeed * currentSpeed);
double b = (2 * targetDirection.Dot(targetLocation - currentLocation) * targetSpeed);
double c = (targetLocation - currentLocation).LengthSquared();

double t;

if (a == 0)
{
t = -c / a;
}
else
{
//Use Formula To Find Root: t = ( -b +- sqrt(b^2 - 4ac) ) / 2a
double u = (b * b) - (4 * a * c);
if (u <= 0)
{
//Root Cannot Be Found. Target Unreachable
return new Vector3D(Double.NaN, Double.NaN, Double.NaN);
}
u = Math.Sqrt(u);

double t1 = (-b + u) / (2 * a);
double t2 = (-b - u) / (2 * a);

t = (t1 > 0 ? (t2 > 0 ? (t1 < t2 ? t1 : t2) : t1) : t2);
}

if (t < 0)
{
return new Vector3D(Double.NaN, Double.NaN, Double.NaN);
}
else
{
return targetLocation + (targetDirection * targetSpeed * t);
}
}

//------------------------------ Missile Aiming Methods ------------------------------

int GetMultiplierSign(double value)
{
return (value < 0 ? -1 : 1);
}

void AimAtTarget()
{
//---------- Activate Gyroscopes To Turn Towards Target ----------

Vector3D yawVector = new Vector3D(targetVector.GetDim(0), 0, targetVector.GetDim(2));
Vector3D pitchVector = new Vector3D(0, targetVector.GetDim(1), targetVector.GetDim(2));
yawVector.Normalize();
pitchVector.Normalize();

targetYawAngle = Math.Acos(yawVector.Dot(Z_VECTOR)) * GetMultiplierSign(targetVector.GetDim(0));
targetPitchAngle = Math.Acos(pitchVector.Dot(Z_VECTOR)) * GetMultiplierSign(targetVector.GetDim(1));

double targetYawAngleLN = Math.Round(targetYawAngle, 2);
double targetPitchAngleLN = Math.Round(targetPitchAngle, 2);

//---------- PID Controller Adjustment ----------

lastYawIntegral = lastYawIntegral + (targetYawAngle / 60);
lastYawIntegral = (INTEGRAL_WINDUP_LIMIT > 0 ? Math.Max(Math.Min(lastYawIntegral, INTEGRAL_WINDUP_LIMIT), -INTEGRAL_WINDUP_LIMIT) : lastYawIntegral);
double yawDerivative = (targetYawAngleLN - lastYawError) * 60;
lastYawError = targetYawAngleLN;
targetYawAngle = (AIM_P * targetYawAngle) + (AIM_I * lastYawIntegral) + (AIM_D * yawDerivative);

lastPitchIntegral = lastPitchIntegral + (targetPitchAngle / 60);
lastPitchIntegral = (INTEGRAL_WINDUP_LIMIT > 0 ? Math.Max(Math.Min(lastPitchIntegral, INTEGRAL_WINDUP_LIMIT), -INTEGRAL_WINDUP_LIMIT) : lastPitchIntegral);
double pitchDerivative = (targetPitchAngleLN - lastPitchError) * 60;
lastPitchError = targetPitchAngleLN;
targetPitchAngle = (AIM_P * targetPitchAngle) + (AIM_I * lastPitchIntegral) + (AIM_D * pitchDerivative);

if (Math.Abs(targetYawAngle) + Math.Abs(targetPitchAngle) > AIM_LIMIT)
{
double adjust = AIM_LIMIT / (Math.Abs(targetYawAngle) + Math.Abs(targetPitchAngle));
targetYawAngle *= adjust;
targetPitchAngle *= adjust;
}

//---------- Set Gyroscope Parameters ----------

SetGyroYaw(targetYawAngle);
SetGyroPitch(targetPitchAngle);
}

void AimAtNaturalGravity()
{
//---------- Activate Gyroscopes To Aim Dampener At Natural Gravity ----------

if (refDownwardBlock == null || naturalGravityLength < 0.01)
{
return;
}

MatrixD dampenerLookAtMatrix = MatrixD.CreateLookAt(POINT_ZERO, refDownwardBlock.WorldMatrix.Forward, (refForwardReverse ? refWorldMatrix.Backward : refWorldMatrix.Forward));

Vector3D gravityVector = Vector3D.TransformNormal(naturalGravity, dampenerLookAtMatrix);
gravityVector.SetDim(1, 0);
gravityVector.Normalize();

if (Double.IsNaN(gravityVector.Sum))
{
gravityVector = new Vector3D(Z_VECTOR);
}

targetRollAngle = Math.Acos(gravityVector.Dot(Z_VECTOR)) * GetMultiplierSign(gravityVector.GetDim(0));

double targetRollAngleLN = Math.Round(targetRollAngle, 2);

//---------- PID Controller Adjustment ----------

lastRollIntegral = lastRollIntegral + (targetRollAngle / 60);
lastRollIntegral = (INTEGRAL_WINDUP_LIMIT > 0 ? Math.Max(Math.Min(lastRollIntegral, INTEGRAL_WINDUP_LIMIT), -INTEGRAL_WINDUP_LIMIT) : lastRollIntegral);
double rollDerivative = (targetRollAngleLN - lastRollError) * 60;
lastRollError = targetRollAngleLN;
targetRollAngle = (AIM_P * targetRollAngle) + (AIM_I * lastRollIntegral) + (AIM_D * rollDerivative);

//---------- Set Gyroscope Parameters ----------

SetGyroRoll(targetRollAngle);
}

//------------------------------ Missile Separation Methods ------------------------------

bool DetachFromGrid()
{
List<IMyTerminalBlock> blocks;
IMyTerminalBlock detachBlock;

switch (missileDetachPortType)
{
case 0:
case 3:
blocks = (strDetachPort != null && strDetachPort.Length > 0 ? GetBlocksWithName<IMyShipMergeBlock>(strDetachPort) : GetBlocksOfType<IMyShipMergeBlock>());
detachBlock = GetClosestBlockFromReference(blocks, Me);

if (detachBlock == null)
{
Echo("Error: Cannot find any Merge Block " + (strDetachPort != null && strDetachPort.Length > 0 ? "with name " + strDetachPort + " to detach" : "to detach."));
return false;
}

detachBlockLocal = detachBlock;
detachBlockRemote = GetConnectedMergeBlock(Me.CubeGrid, detachBlock) as IMyShipMergeBlock;

if (detachBlockRemote == null)
{
Echo("Warning: Merge Block to be detached is not properly connected to another Merge Block (May be due to modded Merge Blocks).");
}

detachBlock.ApplyAction("OnOff_Off");
return true;
case 1:
case 4:
blocks = (strDetachPort != null && strDetachPort.Length > 0 ? GetBlocksWithName<IMyMotorStator>(strDetachPort) : GetBlocksOfType<IMyMotorStator>());
detachBlock = GetClosestBlockFromReference(blocks, Me);

if (detachBlock == null)
{
Echo("Error: Cannot find any Rotor " + (strDetachPort != null && strDetachPort.Length > 0 ? "with name " + strDetachPort + " to detach" : "to detach."));
return false;
}

detachBlockLocal = Me;
detachBlockRemote = detachBlock;

detachBlock.ApplyAction("Detach");
return true;
case 2:
blocks = (strDetachPort != null && strDetachPort.Length > 0 ? GetBlocksWithName<IMyShipConnector>(strDetachPort) : GetBlocksOfType<IMyShipConnector>());
detachBlock = GetClosestBlockFromReference(blocks, Me, true);

if (detachBlock == null)
{
Echo("Error: Cannot find any Connector " + (strDetachPort != null && strDetachPort.Length > 0 ? "with name " + strDetachPort + " to detach" : "to detach."));
return false;
}

detachBlockLocal = detachBlock;
detachBlockRemote = ((IMyShipConnector)detachBlock).OtherConnector;

if (detachBlockRemote == null)
{
Echo("Warning: Connector to be detached is not properly connected to another Connector.");
}

detachBlock.ApplyAction("Unlock");
return true;
case 99:
return true;
default:
Echo("Error: Unknown missileDetachPortType - " + missileDetachPortType + ".");
return false;
}
}

IMyTerminalBlock GetClosestBlockFromReference(List<IMyTerminalBlock> checkBlocks, IMyTerminalBlock referenceBlock, bool sameGridCheck = false)
{
IMyTerminalBlock checkBlock = null;
double prevCheckDistance = Double.MaxValue;

for (int i = 0; i < checkBlocks.Count; i++)
{
if (!sameGridCheck || checkBlocks[i].CubeGrid == referenceBlock.CubeGrid)
{
double currCheckDistance = (checkBlocks[i].GetPosition() - referenceBlock.GetPosition()).Length();
if (currCheckDistance < prevCheckDistance)
{
prevCheckDistance = currCheckDistance;
checkBlock = checkBlocks[i];
}
}
}

return checkBlock;
}

IMyShipMergeBlock GetConnectedMergeBlock(IMyCubeGrid grid, IMyTerminalBlock mergeBlock)
{
IMySlimBlock slimBlock = grid.GetCubeBlock(mergeBlock.Position - new Vector3I(Base6Directions.GetVector(mergeBlock.Orientation.Left)));
return (slimBlock == null ? null : slimBlock.FatBlock as IMyShipMergeBlock);
}

void DetachLockedConnectors()
{
List<IMyTerminalBlock> blocks = GetBlocksOfType<IMyShipConnector>();
for (int i = 0; i < blocks.Count; i++)
{
if (blocks[i].CubeGrid == Me.CubeGrid)
{
IMyShipConnector otherConnector = ((IMyShipConnector)blocks[i]).OtherConnector;
if (otherConnector == null || blocks[i].CubeGrid != otherConnector.CubeGrid)
{
blocks[i].ApplyAction("Unlock");
}
}
}
}

//------------------------------ String Parsing Methods ------------------------------

bool ParseCoordinates(string coordinates, out Vector3D parsedVector)
{
parsedVector = new Vector3D();
coordinates = coordinates.Trim();

double result;
string[] tokens = coordinates.Split(':');

if (coordinates.StartsWith("GPS") && tokens.Length >= 5)
{
if (Double.TryParse(tokens[2], out result))
{
parsedVector.SetDim(0, result);
}
else
{
return false;
}

if (Double.TryParse(tokens[3], out result))
{
parsedVector.SetDim(1, result);
}
else
{
return false;
}

if (Double.TryParse(tokens[4], out result))
{
parsedVector.SetDim(2, result);
}
else
{
return false;
}

return true;
}
else if (coordinates.StartsWith("[T:") && tokens.Length >= 4)
{
if (Double.TryParse(tokens[1], out result))
{
parsedVector.SetDim(0, result);
}
else
{
return false;
}

if (Double.TryParse(tokens[2], out result))
{
parsedVector.SetDim(1, result);
}
else
{
return false;
}

if (Double.TryParse(tokens[3].Substring(0, tokens[3].Length - 1), out result))
{
parsedVector.SetDim(2, result);
}
else
{
return false;
}

return true;
}
else
{
return false;
}
}

//------------------------------ Command Processing Methods ------------------------------

void ProcessConfigurationCommand(string commandLine)
{
string[] keyValues = commandLine.Split(',');

for (int i = 0; i < keyValues.Length; i++)
{
string[] tokens = keyValues[i].Trim().Split(':');
if (tokens.Length > 0)
{
ProcessSingleConfigCommand(tokens);
}
}
}

void ProcessSingleConfigCommand(string[] tokens)
{
string cmdToken = tokens[0].Trim();
if (cmdToken.Equals("MODE") && tokens.Length >= 2)
{
int modeValue;
if (Int32.TryParse(tokens[1], out modeValue))
{
missileLaunchType = modeValue;
}
}
else if (cmdToken.Equals("R_REM") && tokens.Length >= 2)
{
strShipRefRemoteControl = tokens[1];
}
else if (cmdToken.Equals("R_TAR") && tokens.Length >= 2)
{
strShipRefTargetPanel = tokens[1];
}
else if (cmdToken.Equals("R_CAM") && tokens.Length >= 2)
{
strShipRefCamera = tokens[1];
}
else if (cmdToken.Equals("V_DVR") && tokens.Length >= 2)
{
double dvrValue;
if (Double.TryParse(tokens[1], out dvrValue))
{
driftVectorReduction = dvrValue;
}
}
else if (cmdToken.Equals("V_LS") && tokens.Length >= 2)
{
double lsValue;
if (Double.TryParse(tokens[1], out lsValue))
{
launchSeconds = lsValue;
}
}
else if (cmdToken.Equals("V_DRIFT") && tokens.Length >= 2)
{
bool driftValue;
if (bool.TryParse(tokens[1], out driftValue))
{
boolDrift = driftValue;
}
}
else if (cmdToken.Equals("V_LEAD") && tokens.Length >= 2)
{
bool leadValue;
if (bool.TryParse(tokens[1], out leadValue))
{
boolLeadTarget = leadValue;
}
}
else if (cmdToken.Equals("V_DAMP") && tokens.Length >= 2)
{
bool dampenerValue;
if (bool.TryParse(tokens[1], out dampenerValue))
{
boolNaturalDampener = dampenerValue;
}
}
else if (cmdToken.Equals("P_VT") && tokens.Length >= 2)
{
bool vtValue;
if (bool.TryParse(tokens[1], out vtValue))
{
verticalTakeoff = vtValue;
}
}
else if (cmdToken.Equals("P_EMR") && tokens.Length >= 2)
{
bool emrValue;
if (bool.TryParse(tokens[1], out emrValue))
{
excludeGFDMissileRadius = emrValue;
}
}
else if (cmdToken.Equals("P_WFT") && tokens.Length >= 2)
{
bool wftValue;
if (bool.TryParse(tokens[1], out wftValue))
{
waitForHomingTrigger = wftValue;
}
}
else if (cmdToken.Equals("P_EMC") && tokens.Length >= 2)
{
bool emcValue;
if (bool.TryParse(tokens[1], out emcValue))
{
enableMissileCommand = emcValue;
}
}
else if (cmdToken.Equals("SPIN") && tokens.Length >= 2)
{
double spinValue;
if (Double.TryParse(tokens[1], out spinValue))
{
spinAmount = (int)spinValue;
}
}
else if (cmdToken.Equals("CHECK"))
{
subMode = 0;

CheckScript();
}
}

void ProcessTriggerCommands()
{
if (rpmTriggerList != null && rpmTriggerList.Count > 0)
{
int i = 0;
while (i < rpmTriggerList.Count)
{
if (rpmTriggerList[i].Key <= rpm)
{
ProcessSingleMissileCommand(rpmTriggerList[i].Value);
rpmTriggerList.RemoveAt(i);
}
else
{
i++;
}
}
}

if (distTriggerList != null && distTriggerList.Count > 0)
{
int i = 0;
while (i < distTriggerList.Count)
{
if (distTriggerList[i].Key >= distToTarget)
{
ProcessSingleMissileCommand(distTriggerList[i].Value);
distTriggerList.RemoveAt(i);
}
else
{
i++;
}
}
}

if (timeTriggerList != null && timeTriggerList.Count > 0)
{
int i = 0;
while (i < timeTriggerList.Count)
{
if (timeTriggerList[i].Key <= clock)
{
ProcessSingleMissileCommand(timeTriggerList[i].Value);
timeTriggerList.RemoveAt(i);
}
else
{
i++;
}
}
}
}

void ProcessMissileCommand(string commandLine)
{
if (commandLine != null)
{
string[] keyValues = commandLine.Split(',');

for (int i = 0; i < keyValues.Length; i++)
{
string[] tokens = keyValues[i].Trim().Split(':');
if (tokens.Length > 0)
{
ProcessSingleMissileCommand(tokens);
}
}
}
}

void ProcessSingleMissileCommand(string[] tokens)
{
string cmdToken = tokens[0].Trim();
if (cmdToken.Equals("ABORT"))
{
ZeroTurnGyro();

mode = 99;
}
else if (cmdToken.Equals("CRUISE"))
{
if (verticalTakeoff)
{
FireThrusters(launchThrusters, false);
FireThrusters(thrusters, true);
}

ResetGyro();
SetGyroOverride(true);

if (spinAmount > 0)
{
SetGyroRoll(spinAmount);
}

lastTargetPosition = targetPosition = GetFlyStraightVector();

subCounter = 0;
subMode = 0;
mode = 14;
}
else if (cmdToken.StartsWith("ACT") && tokens.Length >= 3)
{
char opCode = (cmdToken.Length >= 4 ? cmdToken[3] : 'B');
List<IMyTerminalBlock> triggerBlocks = null;
switch (opCode)
{
case 'B':
triggerBlocks = GetBlocksWithName(tokens[1]);
break;
case 'P':
triggerBlocks = GetBlocksWithPrefix<IMyTerminalBlock>(tokens[1]);
break;
case 'S':
triggerBlocks = GetBlocksWithSuffix<IMyTerminalBlock>(tokens[1]);
break;
case 'W':
triggerBlocks = GetBlocksContaining<IMyTerminalBlock>(tokens[1]);
break;
}

if (triggerBlocks != null)
{
for (int i = 0; i < triggerBlocks.Count; i++)
{
ITerminalAction action = triggerBlocks[i].GetActionWithName(tokens[2]);
if (action != null)
{
action.Apply(triggerBlocks[i]);
}
}
}
}
else if (cmdToken.StartsWith("SET") && tokens.Length >= 3)
{
char opCode = (cmdToken.Length >= 4 ? cmdToken[3] : 'B');
List<IMyTerminalBlock> triggerBlocks = null;
switch (opCode)
{
case 'B':
triggerBlocks = GetBlocksWithName(tokens[1]);
break;
case 'P':
triggerBlocks = GetBlocksWithPrefix<IMyTerminalBlock>(tokens[1]);
break;
case 'S':
triggerBlocks = GetBlocksWithSuffix<IMyTerminalBlock>(tokens[1]);
break;
case 'W':
triggerBlocks = GetBlocksContaining<IMyTerminalBlock>(tokens[1]);
break;
}

char propCode = (cmdToken.Length >= 5 ? cmdToken[4] : 'P');

if (triggerBlocks != null)
{
for (int i = 0; i < triggerBlocks.Count; i++)
{
switch (propCode)
{
case 'P':
triggerBlocks[i].SetValueFloat(tokens[2], float.Parse(tokens[3]));
break;
case 'B':
triggerBlocks[i].SetValueBool(tokens[2], bool.Parse(tokens[3]));
break;
case 'D':
triggerBlocks[i].SetValueFloat(tokens[2], (float)distToTarget / float.Parse(tokens[3]));
break;
case 'S':
triggerBlocks[i].SetValueFloat(tokens[2], (float)speed / float.Parse(tokens[3]));
break;
case 'T':
triggerBlocks[i].SetValueFloat(tokens[2], (float)(distToTarget / speed) / float.Parse(tokens[3]));
break;
case 'A':
triggerBlocks[i].SetValueFloat(tokens[2], triggerBlocks[i].GetValueFloat(tokens[2]) + float.Parse(tokens[3]));
break;
case 'M':
triggerBlocks[i].SetValueFloat(tokens[2], triggerBlocks[i].GetValueFloat(tokens[2]) * float.Parse(tokens[3]));
break;
}
}
}
}
else if (cmdToken.Equals("SPIN") && tokens.Length >= 1)
{
SetGyroRoll(tokens.Length >= 2 ? Int32.Parse(tokens[1]) : 30);
}
}

void ExecuteTriggerCommand(string commandLine)
{
int startIndex = commandLine.IndexOf('[') + 1;
int endIndex = commandLine.LastIndexOf(']');

string command = (startIndex > 0 && endIndex > -1 ? commandLine.Substring(startIndex, endIndex - startIndex) : commandLine);
string[] keyValues = command.Split(',');

for (int i = 0; i < keyValues.Length; i++)
{
string[] tokens = keyValues[i].Trim().Split(':');
if (tokens.Length > 0)
{
string cmdToken = tokens[0].Trim();
if (cmdToken.Equals("TGR") && tokens.Length >= 3)
{
if (rpmTriggerList == null)
{
rpmTriggerList = new List<KeyValuePair<double, string[]>>();
}

string[] items = new string[tokens.Length - 2];
Array.Copy(tokens, 2, items, 0, items.Length);
rpmTriggerList.Add(new KeyValuePair<double, string[]>(Double.Parse(tokens[1]), items));

haveTriggerCommands = true;
}
else if (cmdToken.Equals("TGD") && tokens.Length >= 3)
{
if (distTriggerList == null)
{
distTriggerList = new List<KeyValuePair<double, string[]>>();
}

string[] items = new string[tokens.Length - 2];
Array.Copy(tokens, 2, items, 0, items.Length);
distTriggerList.Add(new KeyValuePair<double, string[]>(Double.Parse(tokens[1]), items));

haveTriggerCommands = true;
}
else if (cmdToken.Equals("TGE") && tokens.Length >= 3)
{
if (distTriggerList == null)
{
distTriggerList = new List<KeyValuePair<double, string[]>>();
}

string[] items = new string[tokens.Length - 2];
Array.Copy(tokens, 2, items, 0, items.Length);
distTriggerList.Add(new KeyValuePair<double, string[]>(distToTarget - Double.Parse(tokens[1]), items));

haveTriggerCommands = true;
}
else if (cmdToken.Equals("TGT") && tokens.Length >= 3)
{
if (timeTriggerList == null)
{
timeTriggerList = new List<KeyValuePair<int, string[]>>();
}

string[] items = new string[tokens.Length - 2];
Array.Copy(tokens, 2, items, 0, items.Length);
int ticks = (int)(Double.Parse(tokens[1]) * 60) + clock;
timeTriggerList.Add(new KeyValuePair<int, string[]>(ticks, items));

haveTriggerCommands = true;
}
else
{
ProcessSingleMissileCommand(tokens);
}
}
}
}

//------------------------------ Script Debugging Methods ------------------------------

void CheckScript()
{
InitLaunchingShipRefBlocks();

if (shipRefRemoteControl != null)
{
Vector3D targetPoint = shipRefRemoteControl.WorldMatrix.Translation + (shipRefRemoteControl.WorldMatrix.Forward * 500);
Vector3D targetPosition = new Vector3D();
Vector3D targetLocation = new Vector3D();
double targetRadius = 0;

if (FindGFDTarget(shipRefRemoteControl, targetPoint, 500f, out targetPosition, out targetRadius, out targetLocation, GFD_BEAM_WIDTH))
{
Echo("Warning: Something is blocking the GFD lock-on beam within 500 meters in front.");
}
else if (targetRadius == -1)
{
Echo("Warning: There are loose grids (e.g. Joined by rotors, pistons, landing gears, connectors etc) that are interfering with the GFD lock-on beam. Please reposition the Remote Control far away from them.");
}
}

Echo("--- End Of Check, Please Recompile Script And Remove CHECK Argument ---");
}

//------------------------------ Initialization Methods ------------------------------

bool InitLaunchingShipRefBlocks()
{
List<IMyTerminalBlock> blocks;

bool requireRemoteControl = false;
bool requireTargetPanel = false;
bool requireCamera = false;

switch (missileLaunchType)
{
case 0:
case 2:
requireRemoteControl = true;
break;
case 3:
case 5:
requireCamera = true;
break;
case 1:
case 4:
requireTargetPanel = true;
break;
case 6:
requireRemoteControl = true;
requireCamera = true;
break;
case 80:
return true;
case 99:
return true;
default:
Echo("Error: Unknown missileLaunchType - " + missileLaunchType + ".");
return false;
}

if (waitForHomingTrigger || enableMissileCommand)
{
requireTargetPanel = true;
}

if (requireRemoteControl)
{
blocks = GetBlocksWithName<IMyRemoteControl>(strShipRefRemoteControl);

if (blocks.Count == 0)
{
Echo("Error: Required Remote Control with name " + strShipRefRemoteControl + " not found.");
return false;
}
else
{
if (blocks.Count > 1)
{
Echo("Warning: More than one Remote Control with name " + strShipRefRemoteControl + " found. Using first one detected.");
}

shipRefRemoteControl = blocks[0] as IMyRemoteControl;
}
}

if (requireTargetPanel)
{
blocks = GetBlocksWithName<IMyTextPanel>(strShipRefTargetPanel);

if (blocks.Count == 0)
{
Echo("Error: Required Text Panel with name " + strShipRefTargetPanel + " not found.");
return false;
}
else
{
if (blocks.Count > 1)
{
Echo("Warning: More than one Text Panel with name " + strShipRefTargetPanel + " found. Using first one detected.");
}

shipRefTargetPanel = blocks[0] as IMyTextPanel;
}
}

if (requireCamera)
{
blocks = GetBlocksWithName<IMyTerminalBlock>(strShipRefCamera);

if (blocks.Count == 0)
{
Echo("Error: Required Camera Or Aiming Block with name " + strShipRefCamera + " not found.");
return false;
}
else
{
if (blocks.Count > 1)
{
Echo("Warning: More than one Camera Or Aiming Block with name " + strShipRefCamera + " found. Using first one detected.");
}

shipRefCamera = blocks[0];
}

cameraIsTurret = ((shipRefCamera as IMyLargeTurretBase) != null);
}

alertBlock = GridTerminalSystem.GetBlockWithName(strLockTriggerBlock);

notMissile = GridTerminalSystem.GetBlockWithName(strShipRefNotMissile);

return true;
}

bool InitMissileBlocks()
{
gyroscopes = GetGyroscopes();
if (gyroscopes == null) return false;

thrusters = GetThrusters();
if (thrusters == null) return false;

remoteControl = GetRemoteControl();
if (remoteControl == null) return false;

bool isFixedDirection = false;

if (strDirectionRefBlock != null && strDirectionRefBlock.Length > 0)
{
refForwardBlock = GridTerminalSystem.GetBlockWithName(strDirectionRefBlock);
isFixedDirection = (refForwardBlock != null);
}

if (spinAmount > 0)
{
boolNaturalDampener = false;
}

if (refForwardBlock == null || boolNaturalDampener == null || boolDrift == null || verticalTakeoff)
{
thrustValues = ComputeMaxThrustValues(thrusters);
}

if (refForwardBlock == null)
{
refForwardBlock = ComputeHighestThrustReference(thrusters, thrustValues);
refForwardReverse = true;
}

refWorldMatrix = refForwardBlock.WorldMatrix;
if (refForwardReverse)
{
refWorldMatrix = MatrixD.CreateWorld(refWorldMatrix.Translation, refWorldMatrix.Backward, refWorldMatrix.Up);
}

InitGyrosAndThrusters(isFixedDirection);
thrustValues = null;

if (boolLeadTarget == null)
{
boolLeadTarget = true;
}

if (strLockOnSensor != null && strLockOnSensor.Length > 0)
{
List<IMyTerminalBlock> sensors = GetBlocksWithName<IMySensorBlock>(strLockOnSensor);
if (sensors.Count > 0)
{
lockOnSensor = sensors[0] as IMySensorBlock;
}
}

if (strStatusDisplayPrefix != null && strStatusDisplayPrefix.Length > 0)
{
List<IMyTerminalBlock> blocks = GetBlocksWithPrefix<IMyTerminalBlock>(strStatusDisplayPrefix);
if (blocks.Count > 0)
{
statusDisplay = blocks[0];

if (statusDisplay.HasAction("OnOff_On"))
{
statusDisplay.ApplyAction("OnOff_On");

IMyRadioAntenna radioAntenna = statusDisplay as IMyRadioAntenna;
if (radioAntenna != null && !radioAntenna.IsBroadcasting)
{
radioAntenna.ApplyAction("EnableBroadCast");
}
}
}
}

return true;
}

void TurnOnBlocks(List<IMyTerminalBlock> blocks)
{
for (int i = 0; i < blocks.Count; i++)
{
blocks[i].ApplyAction("OnOff_On");
}
}

List<IMyTerminalBlock> GetGyroscopes()
{
List<IMyTerminalBlock> blocks = GetBlocksWithPrefix<IMyGyro>(strGyroscopesPrefix);
if (blocks.Count > 0)
{
return blocks;
}

GridTerminalSystem.GetBlocksOfType<IMyGyro>(blocks);
if (blocks.Count > 0)
{
return blocks;
}
else
{
Echo("Error: No Gyroscopes found.");
return null;
}
}

List<IMyTerminalBlock> GetThrusters()
{
List<IMyTerminalBlock> blocks = GetBlocksWithPrefix<IMyThrust>(strThrustersPrefix);
if (blocks.Count > 0)
{
return blocks;
}

GridTerminalSystem.GetBlocksOfType<IMyThrust>(blocks);
if (blocks.Count > 0)
{
return blocks;
}
else
{
Echo("Error: No Thrusters found.");
return null;
}
}

IMyRemoteControl GetRemoteControl()
{
List<IMyTerminalBlock> blocks = GetBlocksOfType<IMyRemoteControl>();
IMyRemoteControl remoteBlock = (blocks.Count > 0 ? blocks[0] as IMyRemoteControl : null);
if (remoteBlock == null)
{
Echo("Error: No Remote Control found.");
}
return remoteBlock;
}

void InitGyrosAndThrusters(bool isFixedDirection)
{
//---------- Find Gyroscope Orientation With Respect To Ship ----------

gyroYawField = new string[gyroscopes.Count];
gyroPitchField = new string[gyroscopes.Count];
gyroYawFactor = new float[gyroscopes.Count];
gyroPitchFactor = new float[gyroscopes.Count];
gyroRollField = new string[gyroscopes.Count];
gyroRollFactor = new float[gyroscopes.Count];

for (int i = 0; i < gyroscopes.Count; i++)
{
Base6Directions.Direction gyroUp = gyroscopes[i].WorldMatrix.GetClosestDirection(refWorldMatrix.Up);
Base6Directions.Direction gyroLeft = gyroscopes[i].WorldMatrix.GetClosestDirection(refWorldMatrix.Left);
Base6Directions.Direction gyroForward = gyroscopes[i].WorldMatrix.GetClosestDirection(refWorldMatrix.Forward);

switch (gyroUp)
{
case Base6Directions.Direction.Up:
gyroYawField[i] = "Yaw";
gyroYawFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Down:
gyroYawField[i] = "Yaw";
gyroYawFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Left:
gyroYawField[i] = "Pitch";
gyroYawFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Right:
gyroYawField[i] = "Pitch";
gyroYawFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Forward:
gyroYawField[i] = "Roll";
gyroYawFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Backward:
gyroYawField[i] = "Roll";
gyroYawFactor[i] = GYRO_FACTOR;
break;
}

switch (gyroLeft)
{
case Base6Directions.Direction.Up:
gyroPitchField[i] = "Yaw";
gyroPitchFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Down:
gyroPitchField[i] = "Yaw";
gyroPitchFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Left:
gyroPitchField[i] = "Pitch";
gyroPitchFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Right:
gyroPitchField[i] = "Pitch";
gyroPitchFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Forward:
gyroPitchField[i] = "Roll";
gyroPitchFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Backward:
gyroPitchField[i] = "Roll";
gyroPitchFactor[i] = GYRO_FACTOR;
break;
}

switch (gyroForward)
{
case Base6Directions.Direction.Up:
gyroRollField[i] = "Yaw";
gyroRollFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Down:
gyroRollField[i] = "Yaw";
gyroRollFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Left:
gyroRollField[i] = "Pitch";
gyroRollFactor[i] = GYRO_FACTOR;
break;
case Base6Directions.Direction.Right:
gyroRollField[i] = "Pitch";
gyroRollFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Forward:
gyroRollField[i] = "Roll";
gyroRollFactor[i] = -GYRO_FACTOR;
break;
case Base6Directions.Direction.Backward:
gyroRollField[i] = "Roll";
gyroRollFactor[i] = GYRO_FACTOR;
break;
}

gyroscopes[i].ApplyAction("OnOff_On");
}

//---------- Check Whether To Use Default PID Values ----------

if (useDefaultPIDValues)
{
if (Me.CubeGrid.ToString().Contains("Large"))
{
AIM_P = DEF_BIG_GRID_P;
AIM_I = DEF_BIG_GRID_I;
AIM_D = DEF_BIG_GRID_D;
}
else
{
AIM_P = DEF_SMALL_GRID_P;
AIM_I = DEF_SMALL_GRID_I;
AIM_D = DEF_SMALL_GRID_D;
AIM_LIMIT *= 2;
}
}

//---------- Find Forward Thrusters ----------

List<IMyTerminalBlock> checkThrusters = thrusters;
thrusters = new List<IMyTerminalBlock>();

if (!isFixedDirection || boolNaturalDampener == null || boolDrift == null || verticalTakeoff)
{
IMyTerminalBlock leftThruster = null;
IMyTerminalBlock rightThruster = null;
IMyTerminalBlock upThruster = null;
IMyTerminalBlock downThruster = null;

float leftThrustTotal = 0;
float rightThrustTotal = 0;
float upThrustTotal = 0;
float downThrustTotal = 0;

for (int i = 0; i < checkThrusters.Count; i++)
{
Base6Directions.Direction thrusterDirection = refWorldMatrix.GetClosestDirection(checkThrusters[i].WorldMatrix.Backward);
switch (thrusterDirection)
{
case Base6Directions.Direction.Forward:
thrusters.Add(checkThrusters[i]);
break;
case Base6Directions.Direction.Left:
leftThruster = checkThrusters[i];
leftThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Right:
rightThruster = checkThrusters[i];
rightThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Up:
upThruster = checkThrusters[i];
upThrustTotal += thrustValues[i];
if (isFixedDirection)
{
refDownwardBlock = upThruster;
}
break;
case Base6Directions.Direction.Down:
downThruster = checkThrusters[i];
downThrustTotal += thrustValues[i];
break;
}

checkThrusters[i].ApplyAction("OnOff_On");
}

float highestThrust = Math.Max(Math.Max(Math.Max(leftThrustTotal, rightThrustTotal), upThrustTotal), downThrustTotal);
if (highestThrust == 0)
{
if (boolNaturalDampener == true)
{
Echo("Warning: Natural Gravity Dampener feature not possible as there are no Downward Thrusters found.");
}
boolNaturalDampener = false;

if (boolDrift == null)
{
boolDrift = true;
}
}
else
{
if (!isFixedDirection)
{
if (leftThrustTotal == highestThrust)
{
refDownwardBlock = leftThruster;
}
else if (rightThrustTotal == highestThrust)
{
refDownwardBlock = rightThruster;
}
else if (upThrustTotal == highestThrust)
{
refDownwardBlock = upThruster;
}
else
{
refDownwardBlock = downThruster;
}
}
boolNaturalDampener = (refDownwardBlock != null);

if (boolDrift == null)
{
float lowestThrust = Math.Min(Math.Min(Math.Min(leftThrustTotal, rightThrustTotal), upThrustTotal), downThrustTotal);
boolDrift = (highestThrust > lowestThrust * 2);
}
}

if (verticalTakeoff && refDownwardBlock != null)
{
launchThrusters = new List<IMyTerminalBlock>();

for (int i = 0; i < checkThrusters.Count; i++)
{
if (refDownwardBlock.WorldMatrix.Forward.Dot(checkThrusters[i].WorldMatrix.Forward) >= 0.9)
{
launchThrusters.Add(checkThrusters[i]);
}
}
}
}
else
{
for (int i = 0; i < checkThrusters.Count; i++)
{
Base6Directions.Direction thrusterDirection = refWorldMatrix.GetClosestDirection(checkThrusters[i].WorldMatrix.Backward);
if (thrusterDirection == Base6Directions.Direction.Forward)
{
thrusters.Add(checkThrusters[i]);
}
else if (boolNaturalDampener == true && thrusterDirection == Base6Directions.Direction.Up)
{
refDownwardBlock = checkThrusters[i];
}

checkThrusters[i].ApplyAction("OnOff_On");
}

if (boolNaturalDampener == true && refDownwardBlock == null)
{
Echo("Warning: Natural Gravity Dampener feature not possible as there are no Downward Thrusters found.");
boolNaturalDampener = false;
}
}
}

float[] ComputeMaxThrustValues(List<IMyTerminalBlock> checkThrusters)
{
float[] thrustValues = new float[checkThrusters.Count];

for (int i = 0; i < checkThrusters.Count; i++)
{
float prevThrustOverride = checkThrusters[i].GetValueFloat("Override");
checkThrusters[i].SetValueFloat("Override", checkThrusters[i].GetMaximum<float>("Override"));
thrustValues[i] = ((IMyThrust)checkThrusters[i]).ThrustOverride;
checkThrusters[i].SetValueFloat("Override", prevThrustOverride);
}

return thrustValues;
}

IMyTerminalBlock ComputeHighestThrustReference(List<IMyTerminalBlock> checkThrusters, float[] thrustValues)
{
IMyTerminalBlock fwdThruster = null;
IMyTerminalBlock bwdThruster = null;
IMyTerminalBlock leftThruster = null;
IMyTerminalBlock rightThruster = null;
IMyTerminalBlock upThruster = null;
IMyTerminalBlock downThruster = null;

float fwdThrustTotal = 0;
float bwdThrustTotal = 0;
float leftThrustTotal = 0;
float rightThrustTotal = 0;
float upThrustTotal = 0;
float downThrustTotal = 0;

for (int i = 0; i < checkThrusters.Count; i++)
{
Base6Directions.Direction thrusterDirection = Me.WorldMatrix.GetClosestDirection(checkThrusters[i].WorldMatrix.Backward);
switch (thrusterDirection)
{
case Base6Directions.Direction.Forward:
fwdThruster = checkThrusters[i];
fwdThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Backward:
bwdThruster = checkThrusters[i];
bwdThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Left:
leftThruster = checkThrusters[i];
leftThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Right:
rightThruster = checkThrusters[i];
rightThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Up:
upThruster = checkThrusters[i];
upThrustTotal += thrustValues[i];
break;
case Base6Directions.Direction.Down:
downThruster = checkThrusters[i];
downThrustTotal += thrustValues[i];
break;
}
}

List<IMyTerminalBlock> highestThrustReferences = new List<IMyTerminalBlock>(2);

float highestThrust = Math.Max(Math.Max(Math.Max(Math.Max(Math.Max(fwdThrustTotal, bwdThrustTotal), leftThrustTotal), rightThrustTotal), upThrustTotal), downThrustTotal);
if (fwdThrustTotal == highestThrust)
{
highestThrustReferences.Add(fwdThruster);
}
if (bwdThrustTotal == highestThrust)
{
highestThrustReferences.Add(bwdThruster);
}
if (leftThrustTotal == highestThrust)
{
highestThrustReferences.Add(leftThruster);
}
if (rightThrustTotal == highestThrust)
{
highestThrustReferences.Add(rightThruster);
}
if (upThrustTotal == highestThrust)
{
highestThrustReferences.Add(upThruster);
}
if (downThrustTotal == highestThrust)
{
highestThrustReferences.Add(downThruster);
}

if (highestThrustReferences.Count == 1)
{
return highestThrustReferences[0];
}
else
{
Vector3D diagonalVector = ComputeBlockGridDiagonalVector(Me);

IMyTerminalBlock closestToLengthRef = highestThrustReferences[0];
double closestToLengthValue = 0;

for (int i = 0; i < highestThrustReferences.Count; i++)
{
double dotLength = Math.Abs(diagonalVector.Dot(highestThrustReferences[i].WorldMatrix.Forward));
if (dotLength > closestToLengthValue)
{
closestToLengthValue = dotLength;
closestToLengthRef = highestThrustReferences[i];
}
}

return closestToLengthRef;
}
}

Vector3D ComputeBlockGridDiagonalVector(IMyTerminalBlock block)
{
IMyCubeGrid cubeGrid = block.CubeGrid;

Vector3D minVector = cubeGrid.GridIntegerToWorld(cubeGrid.Min);
Vector3D maxVector = cubeGrid.GridIntegerToWorld(cubeGrid.Max);

return (minVector - maxVector);
}

Vector3D ComputeBlockGridMidPoint(IMyTerminalBlock block)
{
return (block.CubeGrid.GridIntegerToWorld(block.CubeGrid.Min) + block.CubeGrid.GridIntegerToWorld(block.CubeGrid.Max)) / 2;
}

//------------------------------ Thruster Control Methods ------------------------------

void FireThrusters(List<IMyTerminalBlock> listThrusters, bool overrideMode)
{
if (listThrusters != null)
{
for (int i = 0; i < listThrusters.Count; i++)
{
listThrusters[i].SetValue<float>("Override", (overrideMode ? listThrusters[i].GetMaximum<float>("Override") : 0f));
}
}
}

//------------------------------ Gyroscope Control Methods ------------------------------

void SetGyroOverride(bool bOverride)
{
for (int i = 0; i < gyroscopes.Count; i++)
{
if (((IMyGyro)gyroscopes[i]).GyroOverride != bOverride)
{
gyroscopes[i].ApplyAction("Override");
}
}
}

void SetGyroYaw(double yawRate)
{
for (int i = 0; i < gyroscopes.Count; i++)
{
gyroscopes[i].SetValueFloat(gyroYawField[i], (float)yawRate * gyroYawFactor[i]);
}
}

void SetGyroPitch(double pitchRate)
{
for (int i = 0; i < gyroscopes.Count; i++)
{
gyroscopes[i].SetValueFloat(gyroPitchField[i], (float)pitchRate * gyroPitchFactor[i]);
}
}

void SetGyroRoll(double rollRate)
{
for (int i = 0; i < gyroscopes.Count; i++)
{
gyroscopes[i].SetValueFloat(gyroRollField[i], (float)rollRate * gyroRollFactor[i]);
}
}

void ZeroTurnGyro()
{
for (int i = 0; i < gyroscopes.Count; i++)
{
gyroscopes[i].SetValueFloat(gyroYawField[i], 0);
gyroscopes[i].SetValueFloat(gyroPitchField[i], 0);
}
}

void ResetGyro()
{
for (int i = 0; i < gyroscopes.Count; i++)
{
gyroscopes[i].SetValueFloat("Yaw", 0);
gyroscopes[i].SetValueFloat("Pitch", 0);
gyroscopes[i].SetValueFloat("Roll", 0);
}
}

//------------------------------ Name Finder API ------------------------------

List<IMyTerminalBlock> GetBlocksOfType<T>() where T: class, IMyTerminalBlock
{
List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.GetBlocksOfType<T>(blocks);

return blocks;
}

List<IMyTerminalBlock> GetBlocksWithName(string name)
{
List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.SearchBlocksOfName(name, blocks);

return blocks;
}

List<IMyTerminalBlock> GetBlocksWithName<T>(string name) where T: class, IMyTerminalBlock
{
nameMatcher = name;

List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.GetBlocksOfType<T>(blocks, MatchNameExact);

return blocks;
}

List<IMyTerminalBlock> GetBlocksWithPrefix<T>(string name) where T: class, IMyTerminalBlock
{
nameMatcher = name;

List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.GetBlocksOfType<T>(blocks, MatchNamePrefix);

return blocks;
}

List<IMyTerminalBlock> GetBlocksWithSuffix<T>(string name) where T: class, IMyTerminalBlock
{
nameMatcher = name;

List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.GetBlocksOfType<T>(blocks, MatchNameSuffix);

return blocks;
}

List<IMyTerminalBlock> GetBlocksContaining<T>(string name) where T: class, IMyTerminalBlock
{
nameMatcher = name;

List<IMyTerminalBlock> blocks = new List<IMyTerminalBlock>();
GridTerminalSystem.GetBlocksOfType<T>(blocks, MatchNameContaining);

return blocks;
}

bool MatchNameExact(IMyTerminalBlock block)
{
return block.CustomName.Equals(nameMatcher);
}

bool MatchNamePrefix(IMyTerminalBlock block)
{
return block.CustomName.StartsWith(nameMatcher);
}

bool MatchNameSuffix(IMyTerminalBlock block)
{
return block.CustomName.EndsWith(nameMatcher);
}

bool MatchNameContaining(IMyTerminalBlock block)
{
return block.CustomName.Contains(nameMatcher);
}